*Universal File Flattener.sas;
*David A. Vandenbroucke, david.a.vandenbroucke@hud.gov, 8/24/2015 2:34 PM; 
TITLE Universal File Flattener;
/*
    This program will produce a single "flat file" SAS dataset from a set of relational
    datasets stored in a SAS library. 
    
    INPUT:  The relational SAS  datasets.  Only the files to be flattened should be
    included in the library.
    
    OUTPUT:
     1. A single SAS file, with one record per case.
     2. A spreadsheet showing some information about the relational input files and the contents
          of the output flat file.
          
     INSTRUCTIONS:
     1. Place all of the relational files in a directory.  NO OTHER SAS DATASETS should 
          be in this directory.  This SHOULD NOT be your SAS WORK library.
     2. Edit the user configuration area below, as follows:
          InName:  Path to the the directory containing the relational files.
          OutName:  Path to the directory where you want the flat file to be saved.
          SpreadDir:  Path to the directory where you want the spreadsheet to be saved.
          DSet:  Name of the flat file SAS dataset (whatever you like).
          Keys:  A space-delimited list of the variables used to identify records in the
               output file.  All records with the same combination of key variables will
               flattened into a single row of the output file.
          OPTIONS:  Edit the SAS system options as you please.
     3. Run this program.  Note that it can take many minutes (but not hours) to run.
     
     CONTACT INFORMATION:
     For questions and bug reports, please contact:

     David A. Vandenbroucke
     Senior Economist
     Office of Policy Development & Research
     U.S. Department of Housing & Urban Development
     451 7th Street SW, Room 8222
     Washington, DC    20410
     
     david.a.vandenbroucke@hud.gov
     phone 202-402-5890
     
*/

/* USER CONFIGURATION AREA. -----------------------------------------------------------------*/

/* Specify the working directories for input and output.  Modify them to fit your installation. 
   These may be the same or different, as you prefer.  DO NOT INCLUDE A FINAL BACKSLASH! 
   The InName path should not be the same as your SAS WORK library. */

%let InName = J:\REH\Dav\data\CURRENT\Conference Notes\2016\SAS Global Forum\DATA\SET1; /* this is the directory where the relational files can be found */
%let OutName = J:\REH\Dav\data\CURRENT\Conference Notes\2016\SAS Global Forum\DATA\Results; /* this is the directory where the flat file will be stored */
%let SpreadDir = J:\REH\Dav\data\CURRENT\Conference Notes\2016\SAS Global Forum\Spreadsheets; /* this is the directory for the contents spreadsheet*/

%LET DSet = Flatfile1;   /* name of flat file.  This can be whatever you want. */
                         
%LET Keys = CKey; /* Variables that identify unique rows in the output file */

OPTIONS   /*General system options.  Modify to fit your preferences */
      NOCENTER NODATE NONUMBER
      LINESIZE=120 PAGESIZE=32627
      COMPRESS=BINARY
      NOFMTERR
      ; * end of options;

/* Macro options.  Use for debugging. */
OPTIONS NOMPRINT NOMLOGIC NOSYMBOLGEN; 

/* END OF USER CONFIGURATION AREA -----------------------------------------------------------*/

/* MACRO DEFINITION AREA --------------------------------------------------------------------*/

/* 
     macro to set all of the macro variables to accomodate the use of the keys.
     The way this macro works is it treats Keys as an array and builds up each of
     the macro variable by addig the keys one at a time.
     
     Note that each declaration is keyed to the point in the program
     where it is used.
*/
%GLOBAL RemoveX AddX XByVars DropVars SQLGroups LastKey; /* make these macro variables global */
%MACRO SetSwitchStrings(Keys);
     
     /* initialize macro variables */
     %LET RemoveX = ;    /* Point 1 */
     %LET AddX = ;       /* Point 2 (used twice) */
     %LET XByVars = ;    /* Point 3 */
     %LET DropVars = ;   /* Point 4 */
     %LET SQLGroups = ;  /* Point 5 */
     %LET Counter = 1;
     %LET OneKey = %SYSFUNC(SCAN(&KEYS.,&Counter.));     /* initialize */
     %DO %UNTIL (&OneKey. = ); /* stops when it runs out of keys */
          %LET RemoveX = &RemoveX. X&OneKey. = &OneKey.;    /* Point 1 */
          %LET AddX = &AddX. &OneKey. = X&Onekey.;          /* Point 2 (used twice) */
          %LET XByVars = &XByVars. X&OneKey.;               /* Point 3 */
          %LET DropVars = &DropVars &OneKey.:;              /* Point 4 */
          %IF &Counter. = 1 %THEN
               %LET SQLGroups = &OneKey.;
          %ELSE     /* We need this to insert the comma delimters */
               %LET SQLGroups = &SQLGroups, &OneKey.;            /* Point 5 */
          %LET LastKey = &OneKey.; /* Keeps just the last pass through the loop */ /* Point 6 */

          /* get the next key to process */          
          %LET Counter = %EVAL(&Counter. + 1);
          %LET OneKey = %SYSFUNC(SCAN(&KEYS.,&Counter.));
      %END;   /* of do until...*/
%MEND;    /*SetSwitchStrings*/

/* macro to flatten one part of the relational file.  These have to be done separately 
     because of the way that PROC TRANSPOSE works.  This macro is called by the FlattenWhole 
     macro below. 
     
     RelFile = name of relational file
     Part = either NUMERIC or CHAR
*/
%MACRO FlattenPart(RelFile,Part); 
     /* First pass:  one record per keyset & variable */
     PROC TRANSPOSE DATA=&RelFile. 
                    OUT=&Part.&RelFile.1
     ;    /* end of proc statement */
        BY &Keys. RecNo;
        VAR _&Part._;
     RUN;

     PROC SORT DATA=&Part.&RelFile.1; 
          BY &Keys. _NAME_ RecNo; 
     RUN;
     
     /* Determine whether the transposed file has a _label_ variable.  This sets a flag
           that will be used in the second pass.  HasLabel = 1 if there are labels and 
           0 if there are not.*/
     PROC SQL NOPRINT;
               SELECT  COUNT(*) INTO :HasLabel FROM 
                         Dictionary.Columns
                              WHERE NAME='_LABEL_' 
                                   AND LIBNAME = 'WORK' 
                                   AND MEMNAME = "&Part.&RelFile.1"
                ;            
     quit; /* SQL */
     run; /* SQL */
           

     /* Second pass:  one record per keyset */
     PROC TRANSPOSE DATA=&Part.&RelFile.1
        OUT=&Part.&RelFile.2 (drop=_name_)
        ; /* end of proc statement */
        BY &Keys.;
        ID _NAME_ RecNo; /* name variables with variable name and number */
        %IF &HasLabel. = 1 %THEN %DO;
               IDLABEL _LABEL_; /* carry along the variable label */
        %END;     /* of if haslabel... */
        VAR COL1;
     run; /* transpose */
%MEND; /* FlattenPart */

/* macro to sort a file by keyset.  This is a macro just to keep the code cleaner. 
     Note that the output is a file in the WORK library.*/
%MACRO SortFile(RelFile);
     PROC SORT DATA= InLib.&RelFile.
          OUT= &RelFile.; 
          BY &Keys.;
     RUN;
%MEND;    /* sortfile */

/* macro to flatten one relational file.  The numeric and character portions are
     flattened separately and then combined.  Note that the input is from the WORK library,
     and that file gets overwritten.
*/     
%MACRO FlattenWhole(RelFile);
     
     DATA &RelFile.; /* number the records in the key set, add to dataset */
          SET &RelFile.;
          BY &Keys.;
          RETAIN RecNo;
          IF first.&LastKey. THEN /* start of a new set of key values */ /* Point 6 */
               RecNo = 1;
          ELSE
               RecNo = RecNo + 1;
     RUN;

     %FlattenPart(&RelFile.,NUMERIC) /*Flatten numeric variables */
     %FlattenPart(&RelFile.,CHAR) /*Flatten character variables */
     
     /*   
          Merge the parts back together. 
          This overwrites the copy in the WORK library.
          The process flattens all of the key variables along with everything else.  This
          data step removes the extraneous  variables.  This requires that the key variables
          be temporarily renamed with an X prefix and then changed back.
     */
     DATA &RelFile. (
                         DROP=RecNo:
                         RENAME = (&RemoveX.)    /* Point 1 */
                    )
          ; /* end of data statement */
          MERGE
               Numeric&RelFile.2 ( RENAME = (&AddX.) ) /* Point 2 */
               Char&RelFile.2 ( RENAME = (&AddX.) )
               ; /* end of merge statement */
          BY &XByVars.;  /* Point 3 */
               
          /* drops all of the extraneous ones (note colon wildcards in DropVars) */
          DROP &DropVars.;    /* Point 4 */
     RUN;     
%MEND;    /* FlattenWhole */ 

/* Macro to determine if the file needs flattening */
%MACRO ProcessFile(RelFile);
     %SortFile(&RelFile)     /* sort input file and save in WORK library */

     /* Determine the maximum number of repeats of the keyset */
     PROC SQL NOPRINT;
          CREATE Table COUNTS AS
               SELECT  COUNT(*) AS counts, &SQLGroups. FROM &RelFile. 
                    GROUP BY &SQLGroups.     /* Point 5 */
                    ; /* end of select count(*)...*/

          SELECT DISTINCT MAX(counts) INTO :MaxCount /* saves the largest value into macro variable */
               FROM COUNTS ; 
     QUIT;
     RUN;
     
     /* Record maximum number of replicates
          NOTE:  This is really ugly.  There has to be a better way that does not involve 
          running a data step for each library member.  Suggestions for improvement would
          cheerfully considered.
     */
     DATA Members;
          SET Members;
          IF Name = "&RelFile." THEN
               MaxReps = &MaxCount.;
     RUN;          
     
     /* flatten if necessary */
     %IF &MaxCount > 1 %THEN
          %FlattenWhole(&Relfile.);
%MEND;    /* ProcessFile */

/* Macro to parse the file list and process each file in turn */
%MACRO ProcessFileList(FileList);
     %LET RelFile = ; /* initialize */
     %LET Counter = 1;
     %LET RelFile = %SYSFUNC(SCAN(&FileList.,&Counter.));
     %DO %UNTIL (&RelFile. = );
         %ProcessFile(&RelFile.); 
          %LET Counter = %EVAL(&Counter. + 1);
          %LET RelFile = %SYSFUNC(SCAN(&FileList.,&Counter.));
       %END;   /* of do until...*/
%MEND;    /* ProcessFileList */

/* END OF MACRO DEFINITION SECTION ----------------------------------------------------------*/

libname InLib "&InName.";    /* input library */
libname FlatLib "&OutName.";  /*output library*/

/* Set macro variables that depend on the keys */
%SetSwitchStrings(&Keys.)

/* Write the list of library members to a dataset */
ODS SELECT NONE;
ODS OUTPUT MEMBERS=Members;
PROC DATASETS LIBRARY=Inlib MEMTYPE=DATA DETAILS;
   RUN;
   QUIT;
ODS OUTPUT CLOSE;
ODS SELECT ALL;

/* build the file list macro variable */
DATA _NULL_;
     Set Members END=last;
     LENGTH FileListVar $ 100;
     RETAIN FileListVar;
     FileListVar = Strip(FileListVar) || " " || Name;
     IF last THEN 
          CALL SYMPUTX("FileList", FileListVar);
RUN;

/* Loop through all the files and flatten the ones that need it */
%ProcessFileList(&FileList.)

/* Merge all the files together */
DATA FlatLib.&DSet.;
     MERGE &FileList.;
     BY &Keys.;
RUN;

/* Write contents to spreadsheet */
ODS TAGSETS.EXCELXP 
     FILE="&SpreadDir.\&DSet. Contents.xml"
     STYLE=JOURNAL2
     OPTIONS(EMBEDDED_TITLES='YES')
     ; * end of ODS;
     
PROC PRINT DATA=MEMBERS NOOBS;
     TITLE2 Information on Relational Datasets Flattened into &DSet.;
RUN;

PROC CONTENTS DATA=FlatLib.&DSet. ;
     TITLE2 Contents of &DSet Flat File;
RUN;

/* Print out the dataset to the spreadsheet.
     THIS IS FOR DEBUGGING PURPOSES ONLY.  Don't try it with a dataset of any size. */
/*PROC PRINT DATA=FlatLib.&DSet.;*/
/*     TITLE2 &DSet Flat File Records;*/
/*RUN;*/
ODS TAGSETS.EXCELXP CLOSE;
